from pathlib import Path

from alembic.config import Config
from alembic import command


def alembic_upgrade_head(host: str, db: str, user: str, password: str | None = None, port: int | str = 5432):
    connect_string = f'postgresql://{user}:{password}@{host}:{port}/{db}'
    root_dir = Path(__file__).parent.parent.parent
    alembic_dir = root_dir / 'rosemary/alembic'
    alembic_cfg = Config()
    alembic_cfg.set_main_option("sqlalchemy.url", connect_string)
    alembic_cfg.set_main_option("script_location", str(alembic_dir))
    command.upgrade(alembic_cfg, "head")



from abc import ABC

from pydantic import BaseModel
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.constants import TypeTaskRosemary
from rosemary.db.models import RosemaryTaskModel
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class InterfaceManualTask(InterfaceRosemaryTask, ABC):
    max_retry = 3
    type_task = TypeTaskRosemary.MANUAL
    timeout = 30

    async def _create_task(self, data: dict, session: AsyncSession, *args, **kwargs):
        new_task = RosemaryTaskModel(
            data=data,
            name=self.__class__.__name__,
            type_task=self.type_task,
            max_retry=self.max_retry,
            timeout=self.timeout
        )
        session.add(new_task)
        await session.commit()
        return new_task.id

    async def create(
            self, *, data: dict | BaseModel | None = None,
            session: AsyncSession | None = None,
            check_exist_repeatable: bool = True
    ):
        data = self._prepare_data_to_db(data)

        if session is None:
            async with self.get_session() as session:
                return await self._create_task(data, session, check_exist_repeatable)
        else:
            return await self._create_task(data, session, check_exist_repeatable)




import asyncio
import inspect
import json
from abc import abstractmethod, ABC

from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModel_V1
from sqlalchemy import select, and_, String, cast
from sqlalchemy.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.constants import TypeTaskRosemary, StatusTaskRosemary
from rosemary.db.models import RosemaryTaskModel
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class InterfaceRepeatableTask(InterfaceRosemaryTask, ABC):
    max_retry = 0
    type_task = TypeTaskRosemary.REPEATABLE
    timeout = 30

    async def _create_task(self, data: dict, session: AsyncSession, check_exist_repeatable: bool = True):
        if check_exist_repeatable:
            for _ in range(self.timeout + 10):
                await asyncio.sleep(1)
                query = select(RosemaryTaskModel).where(and_(
                    RosemaryTaskModel.name == self.get_name(),
                    cast(RosemaryTaskModel.data, String) == json.dumps(data)
                )).where(
                    RosemaryTaskModel.status.in_([
                        StatusTaskRosemary.IN_PROGRESS.value,
                        StatusTaskRosemary.FAILED.value,
                        StatusTaskRosemary.NEW.value,
                    ])
                )
                result = await session.execute(query)
                try:
                    task_db: RosemaryTaskModel = result.scalars().one()
                except NoResultFound:
                    continue
                except MultipleResultsFound:
                    # TODO !!!
                    return
                if task_db:
                    return task_db.id
        new_task = RosemaryTaskModel(
            data=data,
            name=self.__class__.__name__,
            type_task=self.type_task,
            max_retry=self.max_retry,
            timeout=self.timeout
        )
        session.add(new_task)
        await session.commit()
        return new_task.id

    def prepare_data_for_run(self, data: dict):
        return data

    def _prepare_data_to_db(self, data) -> dict | None:
        if data is None:
            return data
        if isinstance(data, dict):
            return data
        if isinstance(data, BaseModel):
            return data.model_dump()
        if isinstance(data, BaseModel_V1):
            return data.dict()
        raise TypeError(f'Incorrect type of data: "{type(data)}"')

    # def create_sync(
    #         self,
    #         *,
    #         data: dict | BaseModel | None = None,
    #         session: AsyncSession | None = None,
    #         check_exist_repeatable: bool = True,
    # ):
    #     return asyncio.run(self.create(
    #         data=data, session=session, check_exist_repeatable=check_exist_repeatable
    #     ))

    async def create(
            self, *, data: dict | BaseModel | None = None,
            session: AsyncSession | None = None,
            check_exist_repeatable: bool = True
    ):
        data = self._prepare_data_to_db(data)

        if session is None:
            async with self.get_session() as session:
                return await self._create_task(data, session, check_exist_repeatable)
        else:
            return await self._create_task(data, session, check_exist_repeatable)

    # async def create(
    #         self,
    #         *,
    #         data: dict | BaseModel | None = None,
    #         session: AsyncSession | None = None,
    #         check_exist_repeatable: bool = True
    # ):
    #
    #     data = self._prepare_data_to_db(data)
    #
    #     if session is None:
    #         async with self.get_session() as session:
    #             asyncio.create_task(self._create_task(data, session, check_exist_repeatable))
    #     else:
    #         asyncio.create_task(self._create_task(data, session, check_exist_repeatable))

    async def prepare_and_run(self, data: dict | None, session: AsyncSession | None):
        prepared_data = self.prepare_data_for_run(data)
        params = inspect.signature(self.run)
        kwargs = {'data': prepared_data}
        for param in params.parameters.keys():
            if param == 'session':
                data['session'] = session
        return await self.run(**kwargs)

    @abstractmethod
    def get_session(self) -> AsyncSession:
        ...

    @abstractmethod
    async def run(self, data):
        ...



import logging


def get_logger(name_worker: str):
    logger = logging.getLogger(f'Rosemary Task -> {name_worker}')
    if logger.handlers:
        return logger
    logger.setLevel(logging.INFO)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)

    logger.addHandler(console_handler)
    return logger



"""init

Revision ID: ddebc53e7879
Revises: 
Create Date: 2023-12-15 21:57:46.846203

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'ddebc53e7879'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('rosemary_tasks',
    sa.Column('id', sa.BIGINT(), autoincrement=True, nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('data', sa.JSON(), nullable=True),
    sa.Column('type_task', sa.Enum('REPEATABLE', 'CRON', 'MANUAL', 'NOT_SETUP', name='typetaskrosemary'), nullable=False),
    sa.Column('status', sa.Enum('NEW', 'IN_PROGRESS', 'FAILED', 'FATAL', 'CANCELED', 'FINISHED', name='statustaskrosemary'), nullable=False),
    sa.Column('retry', sa.Integer(), nullable=False),
    sa.Column('max_retry', sa.Integer(), nullable=False),
    sa.Column('error', sa.String(), nullable=True),
    sa.Column('task_return', sa.String(), nullable=True),
    sa.Column('worker', sa.UUID(), nullable=True),
    sa.Column('timeout', sa.Integer(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_rosemary_tasks_id'), 'rosemary_tasks', ['id'], unique=False)
    op.create_table('rosemary_worker',
    sa.Column('id', sa.BIGINT(), autoincrement=True, nullable=False),
    sa.Column('uuid', sa.UUID(), nullable=False),
    sa.Column('status', sa.Enum('WORKING', 'KILLED', 'CHECKING', name='statusworkerrosemary'), nullable=False),
    sa.Column('ping_time', sa.DateTime(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uuid')
    )
    op.create_index(op.f('ix_rosemary_worker_id'), 'rosemary_worker', ['id'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_rosemary_worker_id'), table_name='rosemary_worker')
    op.drop_table('rosemary_worker')
    op.drop_index(op.f('ix_rosemary_tasks_id'), table_name='rosemary_tasks')
    op.drop_table('rosemary_tasks')
    # ### end Alembic commands ###



import asyncio
import inspect
from abc import abstractmethod, ABC

from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModel_V1
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.constants import TypeTaskRosemary


class InterfaceRosemaryTask(ABC):
    max_retry = 3
    delay_retry = 5  # in sec

    type_task = TypeTaskRosemary.NOT_SETUP
    timeout = 30

    def __init__(self):
        if isinstance(self.type_task, TypeTaskRosemary):
            self._type_task = self.type_task.value
            return

        if isinstance(self.type_task, str):
            if self.type_task in TypeTaskRosemary.values():
                self._type_task = self.type_task
                return
            raise ValueError(
                f'Incorrect value of type task. Expected one of {TypeTaskRosemary.values()}. Got"{self.type_task}"'
            )
        raise TypeError(f'Incorrect type of type task. Expected str or TypeTask. Got "{type(self.type_task)}"')

    @classmethod
    def _is_event_loop_running(cls):
        try:
            loop = asyncio.get_running_loop()
            return True
        except RuntimeError:
            return False

    def cron_setup(self, *, hour: int | str ='*', minutes: int | str = '*', days: str = '*'):
        ...

    def get_name(self) -> str:
        return self.__class__.__name__

    def get_type(self) -> str:
        return self._type_task

    @abstractmethod
    async def _create_task(self, data: dict, session: AsyncSession, check_exist_repeatable: bool = True):
        ...

    def prepare_data_for_run(self, data: dict):
        return data

    def _prepare_data_to_db(self, data) -> dict | None:
        if data is None:
            return data
        if isinstance(data, dict):
            return data
        if isinstance(data, BaseModel):
            return data.model_dump()
        if isinstance(data, BaseModel_V1):
            return data.dict()
        raise TypeError(f'Incorrect type of data: "{type(data)}"')

    # def create_sync(
    #         self,
    #         *,
    #         data: dict | BaseModel | None = None,
    #         session: AsyncSession | None = None,
    #         check_exist_repeatable: bool = True,
    # ):
    #     return asyncio.run(self.create(
    #         data=data, session=session, check_exist_repeatable=check_exist_repeatable
    #     ))

    async def create(
            self, *, data: dict | BaseModel | None = None,
            session: AsyncSession | None = None,
            check_exist_repeatable: bool = True
    ):
        data = self._prepare_data_to_db(data)

        if session is None:
            async with self.get_session() as session:
                return await self._create_task(data, session, check_exist_repeatable)
        else:
            return await self._create_task(data, session, check_exist_repeatable)

    # async def create(
    #         self,
    #         *,
    #         data: dict | BaseModel | None = None,
    #         session: AsyncSession | None = None,
    #         check_exist_repeatable: bool = True
    # ):
    #
    #     data = self._prepare_data_to_db(data)
    #
    #     if session is None:
    #         async with self.get_session() as session:
    #             asyncio.create_task(self._create_task(data, session, check_exist_repeatable))
    #     else:
    #         asyncio.create_task(self._create_task(data, session, check_exist_repeatable))

    async def prepare_and_run(self, data: dict | None, session: AsyncSession | None):
        params = inspect.signature(self.run)
        kwargs = {}
        for param in params.parameters.keys():
            if param == 'data':
                prepared_data = self.prepare_data_for_run(data)
                kwargs['data'] = prepared_data
            if param == 'session':
                kwargs['session'] = session
        return await self.run(**kwargs)

    @abstractmethod
    def get_session(self) -> AsyncSession:
        ...

    @abstractmethod
    async def run(self, data):
        ...
